"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Hostable = void 0;
const protocol_1 = require("@skeldjs/protocol");
const util_1 = require("@skeldjs/util");
const constant_1 = require("@skeldjs/constant");
const events_1 = require("@skeldjs/events");
const objects_1 = require("./objects");
const Networkable_1 = require("./Networkable");
const PlayerData_1 = require("./PlayerData");
const events_2 = require("./events");
/**
 * Represents an object capable of hosting games.
 *
 * See {@link HostableEvents} for events to listen to.
 */
class Hostable extends events_1.EventEmitter {
    constructor(options = {}) {
        super();
        this.options = options;
        this.spawnPrefabs = [
            [objects_1.SkeldShipStatus],
            [objects_1.MeetingHud],
            [objects_1.LobbyBehaviour],
            [objects_1.GameData, objects_1.VoteBanSystem],
            [objects_1.PlayerControl, objects_1.PlayerPhysics, objects_1.CustomNetworkTransform],
            [objects_1.MiraShipStatus],
            [objects_1.PolusShipStatus],
            [objects_1.AprilShipStatus],
            [objects_1.AirshipStatus],
        ];
        this.code = 0;
        this.hostid = -1;
        this.counter = -1;
        this.privacy = "private";
        this.settings = new protocol_1.GameSettings;
        this.objectList = [];
        this.players = new Map;
        this.netobjects = new Map;
        this.registeredPrefabs = new Map;
        this.stream = [];
        this.room = this;
        this.decoder = new protocol_1.PacketDecoder;
        this.shipstatus = undefined;
        this.meetinghud = undefined;
        this.lobbybehaviour = undefined;
        this.gamedata = undefined;
        this.votebansystem = undefined;
        this._incr_netid = 0;
        this._destroyed = false;
        this._started = false;
        this.last_fixed_update = 0;
        if (options.doFixedUpdate) {
            this._interval = setInterval(() => this.FixedUpdate(), Hostable.FixedUpdateInterval);
        }
    }
    destroy() {
        if (this._interval)
            clearInterval(this._interval);
        this._destroyed = true;
    }
    getNextNetId() {
        this._incr_netid++;
        return this._incr_netid;
    }
    /**
     * The current client in the room.
     */
    get myPlayer() {
        return undefined;
    }
    /**
     * The host of the room.
     */
    get host() {
        return this.players.get(this.hostid);
    }
    /**
     * Whether or not a game has started.
     */
    get started() {
        return this._started;
    }
    get destroyed() {
        return this._destroyed;
    }
    /**
     * Whether or not the current client is the host of the room.
     */
    get amhost() {
        return false;
    }
    async broadcast(
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    messages, 
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    reliable = true, 
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    recipient = undefined, 
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    payloads = []
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    ) { }
    async FixedUpdate() {
        var _a;
        const delta = Date.now() - this.last_fixed_update;
        this.last_fixed_update = Date.now();
        for (const [, component] of this.netobjects) {
            if (component &&
                (component.ownerid === ((_a = this.myPlayer) === null || _a === void 0 ? void 0 : _a.clientId) || this.amhost)) {
                component.FixedUpdate(delta / 1000);
                if (component.dirtyBit) {
                    component.PreSerialize();
                    const writer = util_1.HazelWriter.alloc(0);
                    if (component.Serialize(writer, false)) {
                        this.stream.push(new protocol_1.DataMessage(component.netid, writer.buffer));
                    }
                    component.dirtyBit = 0;
                }
            }
        }
        const ev = await this.emit(new events_2.RoomFixedUpdateEvent(this, this.stream));
        if (this.stream.length) {
            const stream = this.stream;
            this.stream = [];
            if (!ev.canceled)
                await this.broadcast(stream);
        }
    }
    /**
     * Resolve a player by some identifier.
     * @param player The identifier to resolve to a player.
     * @returns The resolved player.
     * @example
     *```typescript
     * // Resolve a player by their clientid.
     * const player = room.resolvePlayer(11013);
     * ```
     */
    resolvePlayer(player) {
        const clientid = this.resolvePlayerClientID(player);
        if (clientid === undefined)
            return undefined;
        return this.players.get(clientid);
    }
    /**
     * Resolve a player ID by some identifier.
     * @param player The identifier to resolve to a player ID.
     * @returns The resolved player ID.
     */
    resolvePlayerId(player) {
        if (typeof player === "undefined") {
            return undefined;
        }
        if (typeof player === "number") {
            return player;
        }
        return player.playerId;
    }
    /**
     * Resolve a clientid by some identifier.
     * @param player The identifier to resolve to a client ID.
     * @returns The resolved client ID.
     */
    resolvePlayerClientID(player) {
        if (typeof player === "undefined") {
            return undefined;
        }
        if (typeof player === "number") {
            return player;
        }
        if (player instanceof Networkable_1.Networkable) {
            return player.ownerid;
        }
        if (player instanceof PlayerData_1.PlayerData) {
            return player.clientId;
        }
        return undefined;
    }
    /**
     * Set the code of the room.
     * @example
     *```typescript
     * room.setCode("ABCDEF");
     * ```
     */
    setCode(code) {
        if (typeof code === "string") {
            return this.setCode(util_1.Code2Int(code));
        }
        this.code = code;
    }
    _setPrivacy(privacy) {
        this.privacy = privacy;
    }
    /**
     * Change the the privacy of the room.
     * @param tag The tag to change.
     * @param value The new value of the tag.
     * @example
     *```typescript
     * room.setAlterGameTag(AlterGameTag.ChangePrivacy, 1); // 0 for private, 1 for public.
     * ```
     */
    async setPrivacy(privacy) {
        const oldPrivacy = this.privacy;
        this._setPrivacy(privacy);
        const ev = await this.emit(new events_2.RoomSetPrivacyEvent(this, undefined, oldPrivacy, privacy));
        this._setPrivacy(ev.alteredPrivacy);
        if (ev.alteredPrivacy !== oldPrivacy) {
            await this.broadcast([], true, undefined, [
                new protocol_1.AlterGameMessage(this.code, constant_1.AlterGameTag.ChangePrivacy, this.privacy === "public" ? 1 : 0),
            ]);
        }
    }
    /**
     * Change the settings of the room. If the host, it will broadcast these changes.
     * @param settings The settings to set to (Can be partial).
     * @example
     *```typescript
     * room.syncSettings({
     *   crewmateVision: 0.5,
     *   votingTime: 120
     * });
     * ```
     */
    setSettings(settings) {
        var _a;
        this.settings.patch(settings);
        if (this.amhost) {
            if ((_a = this.myPlayer) === null || _a === void 0 ? void 0 : _a.control) {
                this.myPlayer.control.syncSettings(this.settings);
            }
        }
    }
    /**
     * Set the host of the room. If the current client is the host, it will conduct required host changes.
     * e.g. Spawning objects if they are not already spawned.
     * @param host The new host of the room.
     */
    async setHost(host) {
        const before = this.hostid;
        const resolved_id = this.resolvePlayerClientID(host);
        if (!resolved_id)
            return;
        this.hostid = resolved_id;
        if (this.amhost) {
            if (!this.lobbybehaviour && !this.started) {
                this.spawnPrefab(constant_1.SpawnType.LobbyBehaviour, -2);
            }
            if (!this.gamedata) {
                this.spawnPrefab(constant_1.SpawnType.GameData, -2);
            }
        }
        if (before !== this.hostid && this.host) {
            await this.host.emit(new events_2.PlayerSetHostEvent(this, this.host));
        }
    }
    /**
     * Handle when a client joins the game.
     * @param clientid The ID of the client that joined the game.
     */
    async handleJoin(clientid) {
        if (this.players.has(clientid))
            return null;
        const player = new PlayerData_1.PlayerData(this, clientid);
        this.players.set(clientid, player);
        player.emit(new events_2.PlayerJoinEvent(this, player));
        return player;
    }
    /**
     * Handle when a client leaves the game.
     * @param resolvable The client that left the game.
     */
    async handleLeave(resolvable) {
        const player = this.resolvePlayer(resolvable);
        if (!player)
            return null;
        if (player.playerId !== undefined) {
            if (this.gamedata && this.gamedata.players.get(player.playerId)) {
                this.gamedata.remove(player.playerId);
            }
            if (this.amhost && this.meetinghud) {
                const leftPlayerState = this.meetinghud.voteStates.get(player.playerId);
                if (leftPlayerState) {
                    leftPlayerState.setDead();
                }
                for (const [, voteState] of this.meetinghud.voteStates) {
                    const voteStatePlayer = voteState.player;
                    if (voteStatePlayer && voteState.votedForId === player.playerId) {
                        this.meetinghud.clearVote(voteStatePlayer);
                    }
                }
            }
        }
        if (this.votebansystem && this.votebansystem.voted.get(player.clientId)) {
            this.votebansystem.voted.delete(player.clientId);
        }
        if (player.control) {
            for (const component of player.control.components) {
                this.despawnComponent(component);
            }
        }
        this.players.delete(player.clientId);
        this.emit(new events_2.PlayerLeaveEvent(this, player));
        return player;
    }
    /**
     * Handle when the game is started.
     */
    async handleStart() {
        var _a, _b, _c, _d;
        if (this._started)
            return;
        this._started = true;
        if (this.amhost) {
            await Promise.all([
                Promise.race([
                    Promise.all([...this.players.values()].map((player) => {
                        if (player.isReady) {
                            return Promise.resolve();
                        }
                        return new Promise((resolve) => {
                            player.once("player.ready", () => {
                                resolve();
                            });
                        });
                    })),
                    util_1.sleep(3000),
                ]),
                (_a = this.myPlayer) === null || _a === void 0 ? void 0 : _a.ready(),
            ]);
            const removes = [];
            for (const [clientid, player] of this.players) {
                if (!player.isReady) {
                    await this.handleLeave(player);
                    removes.push(clientid);
                }
            }
            if (removes.length) {
                await this.broadcast([], true, undefined, removes.map((clientid) => {
                    return new protocol_1.RemovePlayerMessage(this.code, clientid, constant_1.DisconnectReason.Error);
                }));
            }
            if (this.lobbybehaviour)
                this.despawnComponent(this.lobbybehaviour);
            const ship_prefabs = [
                constant_1.SpawnType.ShipStatus,
                constant_1.SpawnType.Headquarters,
                constant_1.SpawnType.PlanetMap,
                constant_1.SpawnType.AprilShipStatus,
                constant_1.SpawnType.Airship
            ];
            await this.emit(new events_2.RoomGameStartEvent(this));
            this.spawnPrefab(ship_prefabs[(_b = this.settings) === null || _b === void 0 ? void 0 : _b.map] || 0, -2);
            await ((_c = this.shipstatus) === null || _c === void 0 ? void 0 : _c.selectImpostors());
            await ((_d = this.shipstatus) === null || _d === void 0 ? void 0 : _d.assignTasks());
        }
        else {
            await this.emit(new events_2.RoomGameStartEvent(this));
            if (this.myPlayer)
                await this.myPlayer.ready();
        }
    }
    async _handleEnd(reason) {
        this._started = false;
        this.players.clear();
        for (const [objid] of this.players) {
            this.players.delete(objid);
        }
        await this.emit(new events_2.RoomGameEndEvent(this, reason));
    }
    /**
     * Handle when the game is ended.
     * @param reason The reason for why the game ended.
     */
    async handleEnd(reason) {
        await this._handleEnd(reason);
    }
    /**
     * Handle a client readying up.
     * @param player The client that readied.
     */
    async handleReady(player) {
        const resolved = this.resolvePlayer(player);
        if (resolved) {
            await resolved.ready();
        }
    }
    /**
     * Spawn a component (Not broadcasted to all clients, see {@link Hostable.spawnPrefab}).
     * @param component The component being spawned.
     * @example
     * ```typescript
     * const meetinghud = new MeetingHud(
     *   this,
     *   this.getNextNetId(),
     *   ownerid,
     *   {
     *     dirtyBit: 0,
     *     states: new Map(),
     *   }
     * );
     *
     * this.spawnComponent(meetinghud);
     * ```
     */
    spawnComponent(component, doAwake = true) {
        if (this.netobjects.get(component.netid)) {
            return;
        }
        if (component instanceof objects_1.LobbyBehaviour) {
            if (!this.lobbybehaviour) {
                this.lobbybehaviour = component; // (??)
            }
        }
        if (component instanceof objects_1.GameData) {
            if (!this.gamedata) {
                this.gamedata = component;
            }
        }
        if (component instanceof objects_1.InnerShipStatus) {
            if (!this.shipstatus) {
                this.shipstatus = component;
            }
        }
        if (component instanceof objects_1.VoteBanSystem) {
            this.votebansystem = component;
        }
        if (component instanceof objects_1.MeetingHud) {
            this.meetinghud = component;
        }
        if (doAwake) {
            component.Awake();
        }
        this.netobjects.set(component.netid, component);
        component.emit(new events_2.ComponentSpawnEvent(this, component));
    }
    _despawnComponent(component) {
        this.netobjects.delete(component.netid);
        if (component.owner instanceof PlayerData_1.PlayerData) {
            if (component.owner.character === component) {
                component.owner.character = undefined;
            }
        }
        if (component === this.lobbybehaviour) {
            this.lobbybehaviour = undefined;
        }
        if (component === this.gamedata) {
            this.gamedata = undefined;
        }
        if (component === this.shipstatus) {
            this.shipstatus = undefined;
        }
        if (component === this.votebansystem) {
            this.votebansystem = undefined;
        }
        if (component === this.meetinghud) {
            this.meetinghud = undefined;
        }
        component.Destroy();
        this.objectList.splice(this.objectList.indexOf(component), 1);
        component.components.splice(component.components.indexOf(component), 1);
        component.emit(new events_2.ComponentDespawnEvent(this, component));
    }
    /**
     * Despawn a component.
     * @param component The component being despawned.
     * @example
     *```typescript
     * room.despawnComponent(room.meetinghud);
     * ```
     */
    despawnComponent(component) {
        this._despawnComponent(component);
        this.stream.push(new protocol_1.DespawnMessage(component.netid));
    }
    /**
     * Get an available player ID.
     * @returns The player ID that was found.
     * @example
     *```typescript
     * // Get an available player ID and add it to the gamedata.
     * const playerId = room.getAvailablePlayerID();
     * room.gamedata.add(playerId);
     * ```
     */
    getAvailablePlayerID() {
        for (let i = 0;; i++) {
            if (!this.getPlayerByPlayerId(i)) {
                return i;
            }
        }
    }
    /**
     * Spawn a prefab of an object.
     * @param spawnType The type of object to spawn.
     * @param owner The owner or ID of the owner of the object to spawn.
     * @returns The object that was spawned.
     * @example
     *```typescript
     * room.spawnPrefab(SpawnType.Player, client.me);
     * ```
     */
    spawnPrefab(spawnType, ownerId, flags, componentData = [], doBroadcast = true, doAwake = true) {
        var _a, _b;
        const _ownerid = ownerId === undefined
            ? -2
            : typeof ownerId === "number" ? ownerId : ownerId.clientId;
        const ownerClient = this.players.get(_ownerid);
        const _flags = flags !== null && flags !== void 0 ? flags : (spawnType === constant_1.SpawnType.Player ? constant_1.SpawnFlag.IsClientCharacter : 0);
        let object;
        const spawnPrefab = this.spawnPrefabs[spawnType];
        for (let i = 0; i < spawnPrefab.length; i++) {
            const component = new spawnPrefab[i](this, spawnType, ((_a = componentData[i]) === null || _a === void 0 ? void 0 : _a.netid) || this.getNextNetId(), _ownerid, _flags, undefined, object);
            this._incr_netid = component.netid;
            if (this.netobjects.get(component.netid))
                continue;
            if (i === 0) {
                object = component;
                this.objectList.push(object);
            }
            if ((_b = componentData[i]) === null || _b === void 0 ? void 0 : _b.data) {
                const reader = util_1.HazelReader.from(componentData[i].data);
                component.Deserialize(reader, true);
            }
            this.spawnComponent(component, doAwake);
            object.components.push(component);
        }
        if (spawnType === constant_1.SpawnType.Player && ownerClient) {
            ownerClient.emit(new events_2.PlayerSpawnEvent(this, ownerClient));
        }
        if (this.amhost && doBroadcast) {
            this.stream.push(new protocol_1.SpawnMessage(spawnType, object.ownerid, _flags, object.components.map((component) => {
                const writer = util_1.HazelWriter.alloc(0);
                writer.write(component, true);
                return new protocol_1.ComponentSpawnData(component.netid, writer.buffer);
            })));
        }
        if ((_flags & constant_1.SpawnFlag.IsClientCharacter) > 0 && ownerClient) {
            if (!ownerClient.character) {
                ownerClient.character = object;
                ownerClient.inScene = true;
            }
        }
        return object;
    }
    registerPrefab(spawnType, components) {
        this.spawnPrefabs[spawnType] = components;
    }
    /**
     * Get a player by their player ID.
     * @param playerId The player ID of the player.
     * @returns The player that was found, or null if they do not exist.
     * @example
     * ```typescript
     * const player = room.getPlayerByPlayerId(1);
     * ```
     */
    getPlayerByPlayerId(playerId) {
        for (const [, player] of this.players) {
            if (player.playerId === playerId)
                return player;
        }
        return undefined;
    }
    /**
     * Get a player by one of their components' netids.
     * @param netid The network ID of the component of the player to search.
     * @returns The player that was found, or null if they do not exist.
     * @example
     * ```typescript
     * const player = room.getPlayerByNetId(34);
     * ```
     */
    getPlayerByNetId(netid) {
        for (const [, player] of this.players) {
            if (!player.control)
                continue;
            if (player.control.components.find(component => component.netid === netid)) {
                return player;
            }
        }
        return undefined;
    }
    _getExistingObjectSpawn() {
        const messages = [];
        for (const object of this.objectList) {
            messages.push(new protocol_1.SpawnMessage(object.spawnType, object.ownerid, object.flags, object.components.map(component => {
                const writer = util_1.HazelWriter.alloc(512);
                writer.write(component, true);
                writer.realloc(writer.cursor);
                return new protocol_1.ComponentSpawnData(component.netid, writer.buffer);
            })));
        }
        return messages;
    }
}
exports.Hostable = Hostable;
/**
 * How often a FixedUpdate should be called.
 */
Hostable.FixedUpdateInterval = 1 / 50;
//# sourceMappingURL=Hostable.js.map